/* server.c -	 a server program that uses the socket interface to tcp
 *		 This server serves multiple clients at a time.
 *		 For each client it creates a child process that
 *		 repeatedly reads a string,
 *		 and then sends it back to the client
 *		 Modified for exs API.
 */

/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <time.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <pthread.h>

#include "utilities.h"
#include "latenexs_common.h"


/* Mutex to guard the exit condition and number of child threads created */
static pthread_mutex_t exit_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Guarded by: exit_mutex */
static int nthreads = 0;

/* Guarded by: exit_mutex */
static int exit_status = 0;

/* Condition: nthreads == 0; guarded by: exit_mutex */
static pthread_cond_t exit_cond = PTHREAD_COND_INITIALIZER;

struct child_params {
	int			client_fd;
	struct sockaddr_storage	client_addr;
	};


static inline void *
our_pthread_exit_status(int errno_value)
{
	pthread_mutex_lock(&exit_mutex);
	if (--nthreads == 0) {
		pthread_cond_signal(&exit_cond);
	}
	if (errno_value) {
		exit_status = -errno_value;
	}
	pthread_mutex_unlock(&exit_mutex);
	return NULL;
}	/* fatal_thread_exit */


static void
our_trace_buffer(exs_event_t *evt)
{
#if !defined(ALLOW_DEBUG)
	(void)evt;
#else
	char *buf = evt->exs_evt_union.exs_evt_xfer.exs_evt_buffer;
	size_t len = evt->exs_evt_union.exs_evt_xfer.exs_evt_length;
	if (get_log_level(LOG_APPLICATION) >= LOG_LEVEL_DEBUG) {
		/* make it a string for printing */
		buf[len] = '\0';
		TRACE_DBG(LOG_APPLICATION,
			"%s server: child thread_id %lu recvd %zu byte char string: \"%.40s\"",
			time_prefix(), pthread_self(), len, buf);
	}
#endif
}	/* our_print_buffer */


static void *
server_thread(void *arg)
{
	int			len, client_fd, i, ret;
	size_t			mess_size;
	int			count1;
	char			text_buf[TEXT_SIZE];
	struct child_params	*param;
	struct sockaddr_storage	name;
	socklen_t		namelen;
	exs_event_t		completion;
	exs_qhandle_t		send_queue;
	exs_qhandle_t		recv_queue;
	exs_mhandle_t		in_handle;
	exs_mhandle_t		out_handle;
	char			*in_mess;
	char			*out_mess;
	struct rusage		start_usage;
	int			secs, usecs;
	int			user_secs, user_usecs;
	int			kernel_secs, kernel_usecs;
#if defined(ALLOW_DEBUG)
	struct sockaddr		*client_addr;
#endif


	param = (struct child_params *)arg;
	client_fd = param->client_fd;

	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu connected on "
			"fd %d", time_prefix(), pthread_self(),
			client_fd);

	if (setup_buffers("server", &in_mess, &in_handle,
						&out_mess, &out_handle)) {
		ret = -errno;
		goto out;
	}

#if defined(ALLOW_DEBUG)
	/* we are now successfully connected to a remote client */
	client_addr = (struct sockaddr *)&param->client_addr;
	if (get_log_level(LOG_APPLICATION) >= LOG_LEVEL_DEBUG) {
		if (client_addr != NULL) {
			/* can we get address info about remote client */
			print_ip_address(stderr,
					(struct sockaddr *)&param->client_addr,
					"remote client",
					"latenexs");
		}
	}
#endif

	/* try out the added exs_getsockname() and exs_getpeername() functions*/
	namelen = sizeof(name);
	if (exs_getsockname(client_fd, (struct sockaddr *)&name, &namelen))
		perror("exs_getsockname");
	print_ip_address(stderr, (struct sockaddr *)&name,
			"server exs_getsockname", "latenexs");

	namelen = sizeof(name);
	if (exs_getpeername(client_fd, (struct sockaddr *)&name, &namelen))
		perror("exs_getpeername");
	print_ip_address(stderr, (struct sockaddr *)&name,
			"server exs_getpeername", "latenexs");

	/* create one exs event queue for sending messages to remote side
	 * and another for receiving replies back from the remote side
	 */
	ret = setup_queue("server", "send_queue", &send_queue, SEND_DEPTH,
							EXS_WAIT_ADAPTIVE);
	if (ret) {
		goto out;
	}
	ret = setup_queue("server", "recv_queue", &recv_queue, RECV_DEPTH,
							EXS_WAIT_ADAPTIVE);
	if (ret) {
		goto out;
	}

	/* repeatedly read a message from a client,
	 * then write it to standard output along with the client's id
	 */
	len = size;
	mess_size = size;
	count1 = count - 1;

        /* wait for client to tell us to "go" */
        len = readpacket(client_fd, text_buf, TEXT_SIZE,
                                        EXS_MHANDLE_UNREGISTERED, recv_queue);
	if (len <= 0) {
		fprintf(stderr,"server: server_thread readpacket %s\n",strerror(errno));
		ret = -errno;
		goto out;
	}
	fprintf(stderr, "server: readpacket \"%s\"\n", text_buf);



	/* set up to receive first "ping" from client */
	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu calling first "
			"exs_recv fd %d in_mess %p len %zu recv_queue %p flags "
			"%0x", time_prefix(), pthread_self(), client_fd,
			in_mess, mess_size, recv_queue, recv_flags);
	if (exs_recv(client_fd, in_mess, mess_size, recv_flags,
				recv_queue, NULL, in_handle) < 0) {
		fprintf(stderr, "%s server exs_recv: %s\n",
					time_prefix(), strerror(errno));
		ret = -errno;
		goto err2;
	}

	/* send an ack back to client to say we're ready */
        strcpy(text_buf, "ok");
	fprintf(stderr, "server: writeblock \"%s\"\n", text_buf);
	len = writeblock(client_fd, text_buf, 3, 0,
					EXS_MHANDLE_UNREGISTERED, send_queue);
	if (len <= 0) {
		fprintf(stderr,"server: server_thread writeblock %s\n",
			strerror(errno));
		ret = -errno;
		goto out;
	}

	/* set up to receive next "ping" from client */
	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu calling"
			" exs_recv fd %d in_mess %p len %zu recv_queue "
			"%p flags %0x", time_prefix(), pthread_self(),
			client_fd, in_mess, mess_size, recv_queue,
			recv_flags);
	if (exs_recv(client_fd, in_mess, mess_size, recv_flags,
				recv_queue, NULL, in_handle) < 0) {
		fprintf(stderr, "%s server exs_recv: %s\n",
					time_prefix(), strerror(errno));
		ret = -errno;
		goto err2;
	}

	/* now wait for first "ping" to actually arrive from client */
	if (exs_qdequeue(recv_queue, &completion, 1, NULL) < 0) {
		fprintf(stderr, "%s server recv exs_qdequeue: %s\n",
					time_prefix(), strerror(errno));
		ret = -errno;
		goto err2;
	}

	TRACE_DBG(LOG_APPLICATION, "server: first ping dequeued, start timers");

	/* now start the timers */
	if( getrusage(RUSAGE_SELF, &start_usage) < 0 )
		{
		printf("server: getrusage start %s\n",
					strerror(errno));
		ret = -errno;
		goto err2;
		}
	if (clock_gettime(CLOCK_EXS, &start_time) < 0)
		{
		fprintf(stderr,
			"%s server: clock_gettime start %s\n",
			time_prefix(), strerror(errno));
		ret = -errno;
		goto err2;
		}

	for (i = 0; i < count1; i++) {

		TRACE_DBG(LOG_APPLICATION,"%s server: starting iteration %d",
				time_prefix(), i);

		if (completion.exs_evt_errno != 0) {
			fprintf(stderr, "%s server recv completion: %s\n",
					time_prefix(),
					strerror(completion.exs_evt_errno));
			ret = -completion.exs_evt_errno;
			break;
		}

		len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;

		TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu "
				"exs_recv returned len %d",
				time_prefix(), pthread_self(), len);
		if (len <= 0) {
			ret = -EIO;
			break;
		}

		our_trace_buffer(&completion);

		/* send "pong" reply to "ping" just received (without event) */
		TRACE_DBG(LOG_APPLICATION, "%s server: send pong reply \"%.40s\","
				" mess_size %zu", time_prefix(),
				in_mess, mess_size);
		if (exs_send(client_fd, in_mess, mess_size, EXS_UNSIGNALED,
				NULL, NULL, in_handle) < 0) {
			fprintf(stderr, "%s server exs_send: %s\n",
						time_prefix(), strerror(errno));
			ret = -errno;
			break;
		}

		/* set up to receive next "ping" from client */
		TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu calling"
				" exs_recv fd %d in_mess %p len %zu recv_queue "
				"%p flags %0x", time_prefix(), pthread_self(),
				client_fd, in_mess, mess_size, recv_queue,
				recv_flags);
		if (exs_recv(client_fd, in_mess, mess_size, recv_flags,
					recv_queue, NULL, in_handle) < 0) {
			fprintf(stderr, "%s server exs_recv: %s\n",
						time_prefix(), strerror(errno));
			ret = -errno;
			break;
		}

		/* now wait for next "ping" to actually arrive from client */
		TRACE_DBG(LOG_APPLICATION, "%s server: wait for next ping",
				time_prefix());
		if (exs_qdequeue(recv_queue, &completion, 1, NULL) < 0) {
			fprintf(stderr, "%s server recv exs_qdequeue: %s\n",
						time_prefix(), strerror(errno));
			ret = -errno;
			break;
		}
	}	/* for loop */

	TRACE_DBG(LOG_APPLICATION, "%s server: starting final iteration %d",
							time_prefix(), i);

	if (completion.exs_evt_errno != 0) {
		fprintf(stderr, "%s server final recv completion: %s\n",
				time_prefix(),
				strerror(completion.exs_evt_errno));
		ret = -completion.exs_evt_errno;
		goto err1;
	}

	len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;

	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu "
			"final exs_recv returned len %d",
			time_prefix(), pthread_self(), len);
	if (len <= 0) {
		fprintf(stderr, "server final recv exs_evt_length %d <= 0\n",
			len);
		ret = -EIO;
		goto err1;
	}

	our_trace_buffer(&completion);

	TRACE_DBG(LOG_APPLICATION, "%s server: sending final pong reply, "
					"iteration %d", time_prefix(), i);

	/* send final "pong" reply to final "ping" just received */
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_send final pong reply, "
				"size %zu, iteration %d", time_prefix(),
				mess_size, i);
	if (exs_send(client_fd, in_mess, mess_size, 0,
				send_queue, NULL, in_handle) < 0) {
		fprintf(stderr, "%s server final exs_send: %s\n",
					time_prefix(), strerror(errno));
		ret = -errno;
		goto err1;
	}

	TRACE_DBG(LOG_APPLICATION, "%s server: waiting for final exs_send, "
				"iteration %d", time_prefix(), i);

	/* wait for final "pong" reply to finish */
	if (exs_qdequeue(send_queue, &completion, 1, NULL) < 0) {
		fprintf(stderr, "%s server send final exs_qdequeue: %s\n",
					time_prefix(), strerror(errno));
		ret = -errno;
		goto err1;
	}

	if (completion.exs_evt_errno != 0) {
		fprintf(stderr, "%s server final send completion: %s\n",
				time_prefix(),
				strerror(completion.exs_evt_errno));
		ret = -completion.exs_evt_errno;
	}

	if (completion.exs_evt_union.exs_evt_xfer.exs_evt_length != mess_size) {
		fprintf(stderr, "server final send exs_evt_length %ld != "
			"mess_size %zu\n",
			completion.exs_evt_union.exs_evt_xfer.exs_evt_length,
			mess_size);
	}

	/* don't forget to count the last message */
	i++;
	/* and signal success to the main thread */
	ret = 0;
err1:

	TRACE_DBG(LOG_APPLICATION, "%s server: stop timers, iteration %d",
				time_prefix(), i);

	/* stop timers immediately after last transaction completes */
	get_relative_time(&start_time, &secs, &usecs, "server");
	get_relative_usage(&start_usage, &user_secs, &user_usecs,
					&kernel_secs, &kernel_usecs, "server");

	print_latency("server", i, secs, usecs);
	fprintf(stderr, "server: %11d.%06d seconds elapsed\n", secs, usecs);
        fprintf(stderr, "server: %11d.%06d seconds user (%.4f%%)\n",
			user_secs, user_usecs,
			percentage(secs, usecs, user_secs, user_usecs));
	fprintf(stderr, "server: %11d.%06d seconds kernel (%.4f%%)\n",
			kernel_secs, kernel_usecs,
			percentage(secs, usecs, kernel_secs, kernel_usecs));

err2:
	if (use_registered_memory) {
		/* deregister the RDMA memory buffers */
		if(exs_mderegister(out_handle, 0)) {
			perror("child exs_mderegister out memory");
			ret = -errno;
		}
		if(exs_mderegister(in_handle, 0)) {
			perror("child exs_mderegister  in memory");
			ret = -errno;
		}
	}

	/* close the connection to the client */
	TRACE_DBG(LOG_APPLICATION,
			"%s server: child pthread_id %lu disconnecting from client",
			time_prefix(), pthread_self());
	if (exs_close(client_fd, 0, recv_queue, NULL) < 0) {
		perror("child exs_close connection to client");
		ret = -errno;
	} else {
		do {
			if (asynch_wait(recv_queue, &completion, NULL) < 0) {
				perror("serverexs: server exs_close asynch_wait");
				ret = -errno;
			}
		} while (completion.exs_evt_type != EXS_EVT_CLOSE);
		client_fd = -1;
	}

	if (exs_qdelete(recv_queue) < 0) {
		perror("child exs_qdelete recv_queue");
		ret = -errno;
	}
	if (exs_qdelete(send_queue) < 0) {
		perror("child exs_qdelete send_queue");
		ret = -errno;
	}

out:
	if (client_fd > 0) {
		exs_blocking_close(client_fd);
	}
	free(param);
	return our_pthread_exit_status(ret);
}	/* server_thread */

static int
setup_new_connection(int listening_fd, exs_qhandle_t accept_queue, struct child_params **param)
{
	exs_acceptaddr_t	address_vector[ADDRESS_ITEMS];
	exs_event_t		accept_vector[ACCEPT_ITEMS];
	struct child_params	*new_param;
	int			nevents;

	if (!(new_param = calloc(1, sizeof(*new_param)))) {
		fprintf(stderr, "%s Could not allocate connection parameters: %s\n",
				time_prefix(), strerror(errno));
		return -1;
	}

	/* set up the address vector so exs_accept will return the
	 * peer address of the new client
	 */
	address_vector[0].exs_addr=(struct sockaddr *)&new_param->client_addr;
	address_vector[0].exs_addrlen = sizeof(new_param->client_addr);
	address_vector[0].exs_ahandle = NULL;

	TRACE_DBG(LOG_APPLICATION,
			"%s server: exs_addr %p, exs_addrlen %d,"
			" exs_ahandle %p", time_prefix(),
			address_vector[0].exs_addr,
			address_vector[0].exs_addrlen,
			address_vector[0].exs_ahandle);

	/* accept a client connection */
	if (exs_accept(listening_fd, address_vector, ADDRESS_ITEMS, 0,
			accept_queue) < 0 ) {
		perror("server: exs_accept");
		return -1;
	}
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_accept returned ok",
			time_prefix());

	/* in exs need to wait for accept to complete asynchronously */
	nevents = exs_qdequeue(accept_queue,accept_vector,ACCEPT_ITEMS,
				NULL);
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_qdequeue nevents = %d "
			"accept_vector[0].exs_evt_type = %d",
			time_prefix(), nevents,
			accept_vector[0].exs_evt_type);
	if (nevents < 0) {
		perror("server: exs_qdequeue");
		return -1;
	}
	if (nevents != 1) {
		fprintf(stderr, "%s server: exs_qdequeue dequeued %d "
				"items, expected 1\n",
				time_prefix(), nevents);
		return -1;
	}
	if (accept_vector[0].exs_evt_type != EXS_EVT_ACCEPT) {
		fprintf(stderr, "%s server: exs_qdequeue dequeued item "
				"type %d, expected %d\n", time_prefix(),
				accept_vector[0].exs_evt_type,
				EXS_EVT_ACCEPT);
		return -1;
	}
	TRACE_DBG(LOG_APPLICATION,"%s server: nevents = %d, exs_evt_type ="
			" EXS_EVT_ACCEPT", time_prefix(), nevents);
	if (accept_vector[0].exs_evt_errno != 0) {
		errno = accept_vector[0].exs_evt_errno;
		perror("server: exs_qdequeue status");
		return -1;
	}
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_evt_errno == 0",
			time_prefix());
	new_param->client_fd = accept_vector[0].exs_evt_union.exs_evt_accept
			.exs_evt_new_socket;
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_evt_new_socket %d, "
			"new_client %p", time_prefix(),
			new_param->client_fd, &new_param->client_addr);

	*param = new_param;
	return 0;
}

int
main(int argc, char *argv[])
{
	int			listening_fd, ret;
	exs_qhandle_t		accept_queue;
	struct sockaddr_storage	server_addr;
	struct child_params	*param;
	pthread_t		kid_id;


	startup(argc, argv, 0);


	/* need to create an event_queue for accepting connects to server */
	ret = setup_queue("server", "accept_queue", &accept_queue,
			ACCEPT_DEPTH, EXS_WAIT_NOTIFY);
	if (ret < 0) {
		exit(EXIT_FAILURE);
	}


	/* establish a server "listening post" */
	TRACE_DBG(LOG_APPLICATION, "%s server: calling openserver to "
			"establish listening post", time_prefix());
	listening_fd = openserver(server_port, server_name, ACCEPT_DEPTH,
				socktype,
				(struct sockaddr *)&server_addr,
				&our_conn_params);
	TRACE_DBG(LOG_APPLICATION, "%s server: openserver returned "
			"listening_fd %d", time_prefix(), listening_fd);
	if (listening_fd < 0)
		exit(EXIT_FAILURE);


	print_ip_address(stderr, (struct sockaddr *)&server_addr,
						"server listening", "latenexs");
	print_attributes(stderr, listening_fd, "server");
	fprintf(stderr, "Waiting for %d clients\n", n_listener_loops);

	/* we are now successfully established as a server */
	while (n_listener_loops-- > 1) {

		if (setup_new_connection(listening_fd, accept_queue, &param) < 0) {
			pthread_mutex_lock(&exit_mutex);
			exit_status = -errno;
			pthread_mutex_unlock(&exit_mutex);
			goto out;
		}
		pthread_mutex_lock(&exit_mutex);
		++nthreads;
		pthread_mutex_unlock(&exit_mutex);

		errno = pthread_create(&kid_id, NULL, server_thread, param);
		if (errno != 0) {
			perror("server: pthread_create");

			pthread_mutex_lock(&exit_mutex);
			--nthreads;
			pthread_mutex_unlock(&exit_mutex);
		} else {
			pthread_detach(kid_id);
			fprintf(stderr, "*** child with thread_id %lu started\n",
				(long)kid_id);
		}
	}

	/* Process last connection synchronously, since this thread can do
	 * nothing else until all other listening threads terminate anyways. */
	if (setup_new_connection(listening_fd, accept_queue, &param) < 0) {
		pthread_mutex_lock(&exit_mutex);
		exit_status = -errno;
		pthread_mutex_unlock(&exit_mutex);
		goto out;
	}
	pthread_mutex_lock(&exit_mutex);
	++nthreads;
	pthread_mutex_unlock(&exit_mutex);
	server_thread(param);

out:
	pthread_mutex_lock(&exit_mutex);
	while (nthreads != 0) {
		pthread_cond_wait(&exit_cond, &exit_mutex);
	}
	pthread_mutex_unlock(&exit_mutex);

	if (exs_qdelete(accept_queue) < 0) {
		perror("server: exs_qdelete");
	}

	/* close the "listening post" socket by which server made connections */
	TRACE_DBG(LOG_APPLICATION, "%s server: closing listening post %d",
			time_prefix(), listening_fd);
	if (exs_blocking_close(listening_fd) < 0) {
		perror("server: exs_blocking_close listening post");
	}

	return exit_status ? EXIT_FAILURE : EXIT_SUCCESS;
}	/* main */
