/* utilities.h */

/* Modified for exs API */

/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef UTILITIES_H
#define UTILITIES_H

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <exs.h>

/* CLOCK_EXS is CLOCK_MONOTONIC if the system supports it; otherwise, it is
 * CLOCK_REALTIME.  This clock is used for performance timings, where elapsed
 * time is more important than absolute time. */
#if defined(_POSIX_MONOTONIC_CLOCK) && _POSIX_MONOTONIC_CLOCK != -1
#define CLOCK_EXS CLOCK_MONOTONIC
#else
#define CLOCK_EXS CLOCK_REALTIME
#endif

#include <time.h>

enum util_log_module {
	LOG_UTILITIES,
	LOG_APPLICATION,
	LOG_MAX,
};

/* These are in descending order of severity, so that default static
 * initialization of the array gives no extraneous trace messages. */
enum util_log_level {
	LOG_LEVEL_NONE,
	LOG_LEVEL_INFO,
	LOG_LEVEL_DEBUG
};

#if !defined(ALLOW_DEBUG)

#define TRACE_DBG(...) do {} while (0);

static inline void
set_log_level(unsigned int module, unsigned int level)
{
	(void)module;
	(void)level;
}

static inline unsigned int
get_log_level(unsigned int module)
{
	(void)module;
	return 0U;
}

#else

#include <stdarg.h>

extern unsigned int exs_util_log_level[];

static inline void
set_log_level(unsigned int module, unsigned int level)
{
	assert(module < LOG_MAX);
	exs_util_log_level[module] = level;
}	/* set_log_level */

static inline unsigned int
get_log_level(unsigned int module)
{
	assert(module < LOG_MAX);
	return exs_util_log_level[module];
}

static inline void __attribute__((format(printf, 4, 5)))
do_trace(const char *file, int line, const char *function,
		const char *format, ...)
{
	va_list args;

	flockfile(stderr);
	fprintf(stderr, "%s:%d %s ",
			file, line, function);
	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	putc_unlocked('\n', stderr);
	funlockfile(stderr);
}

#define TRACE_DBG(module, ...) \
	if (get_log_level((module)) >= LOG_LEVEL_DEBUG) { \
		do_trace(__FILE__, __LINE__, __func__, __VA_ARGS__); \
	}

#define TRACE_INFO(module, ...) \
	if (get_log_level((module)) >= LOG_LEVEL_INFO) { \
		do_trace(__FILE__, __LINE__, __func__, __VA_ARGS__); \
	}

#endif

/* EXS connection parameters. Value parameters may be left at their default by
 * using a value of INT_MAX here. */
struct conn_params {
	/* all fd flags for fcntl OR'd together */
	int fd_flags;

	/* Small packet size for Nagle */
	int small_packet_max_size;

	/* Maximum amount of inline data to attempt to use */
	int max_inline;

	/* CPU to bind completion thread to */
	int comp_cpu;

	/* Maximum number of flow control credits */
	int max_credits;

	/* SOCK_STREAM buffer size */
	size_t stream_buf_size;
};

extern void no_sigpipe(void);

extern void fstatdump( int limit );

void blockio_set_debug(int flag);

/* Returns after starting write of nbytes characters from buffer to fd.  */
int
asynch_send(int fd, char *buffer, int nbytes, int flags,
		exs_mhandle_t buf_handle, exs_qhandle_t wait_queue);

/* Returns after starting write of count vectors from sendvec to fd.  */
int
asynch_sendfile(int fd, exs_xferfile_t *sendvec, int count,
		exs_qhandle_t wait_queue, exs_ahandle_t ahandle);

/* Returns after starting read of nbytes characters into buffer from fd.  */
int
asynch_recv(int fd, char *buffer, int nbytes, int flags,
		exs_qhandle_t wait_queue,
		exs_ahandle_t ahandle,
		exs_mhandle_t buf_handle);

/* Returns after getting an async event from wait_queue	*/
int
asynch_wait(exs_qhandle_t wait_queue, exs_event_t *vector,
		struct timeval *timeout);

int writeblock(int fd, char *buffer, int nbytes, int flags,
		exs_mhandle_t buf_handle,
		exs_qhandle_t wait_queue);

int readblock(int fd, char *buffer, int nbytes, exs_qhandle_t wait_queue);

/* Returns after <= nbytes characters have been read into buffer from fd.
 * Will also return if an end of file is encountered.
 * Value returned is actual number of bytes read.
 */
int readpacket(int fd, char *buffer, int nbytes,
		exs_mhandle_t buf_handle,
		exs_qhandle_t wait_queue);

int print_attributes(FILE *fp, int fd, char *message);

/* convert and print IP address as dotted-decimal or colon-hex string */
void
print_ip_address(FILE *fp, struct sockaddr *addr, char *who, char *message);

int openclient(char *server_port, char *server_node, int socktype,
		struct sockaddr *server_addr,
		struct sockaddr *client_addr,
		const struct conn_params *params);

int openserver(char *server_port, char *server_name, int listening_depth,
		int socktype, struct sockaddr *server_address,
		const struct conn_params *params);

int
setup_queue(const char *who, const char *name, exs_qhandle_t *queue, int depth,
								int wait_type);

double
percentage(int secs, int usecs, int top_secs, int top_usecs);

void
get_relative_time(struct timespec *start_time, int *secs, int *usecs,
							const char *message);

void
get_relative_usage(struct rusage *start_usage, int *user_secs, int *user_usecs,
		int *kernel_secs, int *kernel_usecs, const char *message);

int
dispatch_exs_events(exs_qhandle_t qh, int max_count,
					int (* const handlers[])(exs_event_t *),
					const struct timeval *timeout);

int
exs_connect_new(const char *const host, const char *const port,
		int family, int socktype, int flags, struct timeval *timeout,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle);

int
our_bind_all(const char *hostname, const char *service,
		int family, int socktype, int **fds, int *nfds);

void
print_name_info(int fd, const struct sockaddr *sa, socklen_t salen,
							const char *message);

#endif
