/* cmdline.c */

/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef _CMDLINE_H
#define _CMDLINE_H

#include "port.h"

#define LOGFILE_DEFAULT "test-suite-exs-full.log"
#define OPTIONS ":a:A:p:PvVhHnNlLo:O:t:T:r:R:"

/* defines how to use the framework at the command line */
static void usage(void)
{
	fprintf(stdout, "Usage: \n"
		"-h -H - help (usage)\n"
		"-l -L - show all tests (with their numbers)\n"
		"-t -T - specify test number to show (0 will show all)\n"
		"-v -V - indicates verbose output\n"
		"-a -A - specify server address (required)\n"
		"-p -P - specify port number (default is 51515)\n"
		"-n -N - specify non-blocking (asynchronous) mode\n"
		"-o -O - specify log file (default is " LOGFILE_DEFAULT ")\n"
		"-r -R - specify timeout window (default is 30 seconds)"
		"0     - specify which test to run\n"
		"        can be separated by spaces\n"
		"0-1   - specify a range of tests\n");
	exit(0);
}

/* scan the command line to find and process all switches */
static int setup_options(int argc, char *argv[], param * p)
{
	char *logfile = NULL;	/* set by -o to name of log file */
	int result;		/* cumulative result of all scanning */
	char c;			/* latest switch char found by getopt */

	/* assume no error */
	result = 0;

	/* default synchronous mode */
	p->mode = 0;

	/* list flag is false by default */
	p->list = 0;

	/* set all non-zero defaults before scanning the command line */
	p->family = PF_INET;
	p->socktype = SOCK_STREAM;
	p->protocol = IPPROTO_TCP;

	p->addr = port_name;
	p->port = port_num;

	//TODO if async, change this
	p->flags = EXS_BLOCK;
	p->close_flags = EXS_BLOCK | EXS_DONTLINGER;
	p->to = NULL;

	/* set up evtvec */
	p->evtvec_cnt = 1;
	exs_event_t evtvec;
	p->evtvec = &evtvec;

	/* set up send buffer */
	p->SEND_BUFSIZE = 100;
	char send_buf[p->SEND_BUFSIZE];
	memset(send_buf, 0, sizeof(*send_buf));
	p->send_buf = send_buf;

	/* set up recv buffer */
	p->RECV_BUFSIZE = 100;
	char recv_buf[p->RECV_BUFSIZE];
	memset(recv_buf, 0, sizeof(*recv_buf));
	p->recv_buf = recv_buf;

	/* default exs_mhandle_t */
	p->mh = EXS_MHANDLE_UNREGISTERED;

	/* initialize linked list for selected tests */
	struct node *root;
	root = malloc(sizeof(struct node));
	root->next = NULL;
	struct node *tail;
	tail = root;
	tail->value = 0;

	/* tell getopt() to NOT print error messages */
	opterr = 0;

	/* loop on each switch in the command line */
	while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (c) {
		case 't':
		case 'T':
			p->list = 1;
			p->list_opt = optarg;
			break;
		case 'l':
		case 'L':
			p->list = 1;
			break;
		case 'h':
		case 'H':
			usage();
			break;
		case 'a':
		case 'A':
			p->addr = optarg;
			break;
		case 'p':
		case 'P':
			p->port = optarg;
			break;
		case 'n':
		case 'N':
			/** asynchronous mode **/
			mode = 1;
			p->mode = 1;
			break;
		case 'r':
		case 'R':
			timeout = atoi(optarg);
			break;
		case 'o':
		case 'O':
			logfile = optarg;
			break;
		case 'v':
		case 'V':
			VERBOSE = 1;
			break;
		default:	/* unimplemented option */
			fprintf(stderr, "ERROR option -%c: %s\n",
				c, "unimplemented");
			result = -1;
			break;
		}		/* switch */
	}			/* while */

    /** if the user did not specify test numbers, run all tests **/

	if (p->mode > 0) {
		//TODO if async, change this
		p->flags = 0;
		p->close_flags = 0;
		p->to = NULL;
	}

    /** if the user did specify test numbers, execute those tests **/
	int i;
	for (i = 0; i < argc - optind; i++) {
		int index = optind + i;
		char *nstr = argv[index];
		int nint = atoi(nstr);

		while (tail->value != 0) {
			tail = tail->next;
		}

		if (strstr(nstr, "-") != NULL) {
			char *right_side = strstr(nstr, "-");
			int str_len = strlen(nstr);
			int right_len = strlen(right_side);
			int offset = str_len - right_len;

			char left_substr[offset];
			memcpy(left_substr, &nstr[0], offset);
			left_substr[offset] = '\0';

			char right_substr[right_len];
			memcpy(right_substr, &nstr[offset + 1],
			       right_len - 1);
			left_substr[right_len - 1] = '\0';

			int l = atoi(left_substr);
			int r = atoi(right_substr);

			if (l >= r) {
				fprintf(stderr,
					"ERROR: Invalid range (%s)\n",
					nstr);
				tail->value = -1;
				tail->next = malloc(sizeof(struct node));
				tail = tail->next;
			} else {
				int i;
				for (i = l; i < r + 1; i++) {
					tail->value = i;
					tail->next =
					    malloc(sizeof(struct node));
					tail = tail->next;
				}
			}
		} else {
			tail->value = nint;
			tail->next = malloc(sizeof(struct node));
		}
	}

	p->iter = root;

	if (open_log_stream(logfile ? logfile : LOGFILE_DEFAULT) != 0) {
		if (!logfile) {
			/* User did not specify a custom log file.  Continue
			 * running with a warning that the test results will not
			 * be logged. */
			fprintf(stderr, "\nWARNING: Could not open default log file %s for writing: %s\n"
				"WARNING: Test results will not be logged!\n\n",
				LOGFILE_DEFAULT, strerror(errno));
		} else {
			/* User specified a custom log file, so if we can't open
			 * it, it's a fatal error. */
			fprintf(stderr, "Could not open log file %s for writing: %s\n",
				logfile, strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	return result;
}				/* setup_options */

#endif
