/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef _RESULT_H
#define _RESULT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>
#include <exs.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>

int VERBOSE = 0;

int mode = 0;

int timeout = 30;

#define TMP_PATH_MAX 100

char* exs_test_temp = "/tmp";
char* exs_test_fifo_1 = "exs_test_fifo1";
char* exs_test_fifo_2 = "exs_test_fifo2";
char exs_test_small_file[TMP_PATH_MAX];
char exs_test_large_file[TMP_PATH_MAX];

#include "inc/fault.h"
#include "inc/param.h"
#include "inc/print.h"
#include "cmdline.h"

static pid_t run_test(param *p, int i, void *test_func)
{
	pid_t pid = fork();
	if (i == -1)
		return pid;	//invalid range provided
	if (pid == 0) {
		alarm(timeout);
		int test_id = i;
		p->test_number = test_id;
		pthread_t test_thread;
		if (pthread_create(&test_thread, NULL, test_func, p)) {
			fprintf(stderr, "Error creating thread\n");
			exit(EXIT_FAILURE);
		}
		if (pthread_join(test_thread, NULL)) {
			fprintf(stderr, "Error joining thread\n");
			exit(EXIT_FAILURE);
		}
		exit(EXIT_SUCCESS);
	} else {
		int returnStatus;
		waitpid(pid, &returnStatus, 0);
	}
	return pid;
}

static void list_tests(int select, char **all_tests, int num_tests)
{
	if (select < 0 || select > num_tests) {
		fprintf(stdout, "Invalid -t test number (1 - %d)\n",
				num_tests);
		exit(0);
	}

	if (select == 0) {
		int i;
		for (i = 0; i < num_tests; i++) {
			if (strcmp(all_tests[i], "") != 0)
				fprintf(stdout, "%-2d : %s\n",
						i + 1, all_tests[i]);
		}
	} else {
		fprintf(stdout, "%d : %s\n", select, all_tests[select - 1]);
	}

	exit(0);
}

static void start_sendfile(void)
{
	int fd, ret;

	fd = open(exs_test_small_file, O_WRONLY|O_CREAT|O_EXCL, 0600);
	if (fd < 0) {
		fprintf(stderr, "Could not open %s for writing: %s\n",
				exs_test_small_file, strerror(errno));
		exit(EXIT_FAILURE);
	}
	ret = ftruncate(fd, 1024);
	if (ret != 0) {
		fprintf(stderr, "cannot truncate %s: %s\n",
				exs_test_small_file,
				strerror(errno));
		exit(EXIT_FAILURE);
	}
	ret = close(fd);
	if (ret != 0) {
		fprintf(stderr, "cannot close %s: %s\n",
				exs_test_small_file,
				strerror(errno));
		exit(EXIT_FAILURE);
	}

	fd = open(exs_test_large_file, O_WRONLY|O_CREAT|O_EXCL, 0600);
	if (fd < 0) {
		fprintf(stderr, "Could not open %s for writing: %s\n",
				exs_test_large_file, strerror(errno));
		exit(EXIT_FAILURE);
	}
	ret = ftruncate(fd, EXS_MAX_SENDFILE_SIZE + 1);
	if (ret != 0) {
		fprintf(stderr, "cannot truncate %s: %s\n",
				exs_test_large_file,
				strerror(errno));
		exit(EXIT_FAILURE);
	}
	ret = close(fd);
	if (ret != 0) {
		fprintf(stderr, "cannot close %s: %s\n",
				exs_test_large_file,
				strerror(errno));
		exit(EXIT_FAILURE);
	}
}

static void end_sendfile(void)
{
	remove(exs_test_small_file);
	remove(exs_test_large_file);
}

static int execute(int argc, char *argv[], void *test_func, char **all_tests, int num_tests)
{
	param *p = calloc(1, sizeof(param));
	p->func = argv[0];	//"test_exs_init";

	/* get temporary directory */
	char* tmpenv = getenv("TMPDIR");
	if( tmpenv != NULL ) exs_test_temp = tmpenv;

	/* create temporary directory */
	char tmp[TMP_PATH_MAX];
	int n = snprintf(tmp, TMP_PATH_MAX, "%s/exs_test_XXXXXX", exs_test_temp);
	if ( n > TMP_PATH_MAX ){
		fprintf(stdout, "FAIL: $TMPDIR too large for buffer; exiting\n");
		return -1;
	}
	char* tmpdir = mkdtemp(tmp);

	/* create fifos */
	char fifo1[TMP_PATH_MAX];
	n = snprintf(fifo1, TMP_PATH_MAX, "%s/%s", tmpdir, exs_test_fifo_1);
	if ( n > TMP_PATH_MAX ){
		fprintf(stdout, "FAIL: $TMPDIR too large for buffer; exiting\n");
		return -1;
	}
	mkfifo(fifo1, 0600);

	char fifo2[TMP_PATH_MAX];
	n = snprintf(fifo2, TMP_PATH_MAX, "%s/%s", tmpdir, exs_test_fifo_2);
	if ( n > TMP_PATH_MAX ){
		fprintf(stdout, "FAIL: $TMPDIR too large for buffer; exiting\n");
		return -1;
	}
	mkfifo(fifo2, 0600);

	n = snprintf(exs_test_small_file, TMP_PATH_MAX, "%s/small.txt", tmpdir);
	if ( n > TMP_PATH_MAX ){
		fprintf(stdout, "FAIL: $TMPDIR too large for buffer; exiting\n");
		return -1;
	}

	n = snprintf(exs_test_large_file, TMP_PATH_MAX, "%s/large.txt", tmpdir);
	if ( n > TMP_PATH_MAX ){
		fprintf(stdout, "FAIL: $TMPDIR too large for buffer; exiting\n");
		return -1;
	}

    /** get command line options **/
	if (setup_options(argc, argv, p) != 0) {
		return 0;
	}

	/** list tests, then end **/
	if (p->list == 1) {
		int opt;
		if (p->list_opt != NULL && !isdigit(p->list_opt[0])) {
			fprintf(stdout, "-t option is not a number\n");
			exit(0);
		}

		if (p->list_opt == NULL)
			opt = 0;
		if (p->list_opt != NULL)
			opt = atoi(p->list_opt);

		list_tests(opt, all_tests, num_tests);
		exit(0);
	}

    /** run all tests (no tests specified) **/
	if (p->iter->value == 0) {
		int i;
		for (i = 1; i < num_tests + 1; i++) {
			p->test_number = i;

			if( strstr(argv[0], "sendfile") != NULL ){
				start_sendfile();
			}

			if (run_test(p, i, test_func) == 0)
				break;

			if( strstr(argv[0], "sendfile") != NULL ){
				end_sendfile();
			}
		}
	}
    /** run specified tests **/
	else {
		struct node *iter = p->iter;
		while (iter->next != NULL) {
			p->test_number = iter->value;

			if( strstr(argv[0], "sendfile") != NULL ){
				start_sendfile();
			}

			if (run_test(p, iter->value, test_func) == 0)
				break;

			if( strstr(argv[0], "sendfile") != NULL ){
				end_sendfile();
			}

			iter = iter->next;
		}
	}

	/* remove files */
	remove(fifo1);
	remove(fifo2);

	/* remove temp directory */
	rmdir(tmpdir);

	return 0;
}
#endif
