/*
 * Copyright 2016-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_init_setup.h"
#include "setup/exs_socket_setup.h"

#define GETSOCKOPT_TESTS 11

static char *all_getsockopt_tests[GETSOCKOPT_TESTS] = {
	"exs_getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &v, &len) [no exs_init]",
	"exs_getsockopt(BAD_FD, SOL_SOCKET, SO_REUSEADDR, &v, &len)",
	"exs_getsockopt(fd, BAD_LEVEL, SO_REUSEADDR, &v, &len)",
	"exs_getsockopt(fd, SOL_SOCKET, BAD_OPTNAME, &v, &len)",
	"exs_getsockopt(fd, SOL_EXS, BAD_OPTNAME, &v, &len)",
	"exs_getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &v, &len)",
	"exs_getsockopt(fd, SOL_EXS, SO_INLINE, &v, &len)",
	"exs_getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &v, &TOO_SMALL_OPTLEN)",
	"exs_getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &v, &TOO_LARGE_OPTLEN)",
	"exs_getsockopt(fd, SOL_EXS, SO_INLINE, &v, &TOO_SMALL_OPTLEN)",
	"exs_getsockopt(fd, SOL_EXS, SO_INLINE, &v, &TOO_LARGE_OPTLEN)",
};

static void test_getsockopt_enoprotoopt(param *p, int level)
{
	socklen_t len;
	int v, ret;

	p->test_value = "ENOPROTOOPT";
	p->test_name = all_getsockopt_tests[p->test_number - 1];
	setup_fault_catcher(p, 0);
	if (ok_exs_init(p) < 0)
		exit(EXIT_FAILURE);
	if (ok_exs_socket(p) < 0)
		exit(EXIT_FAILURE);
	len = sizeof(v);
	ret = exs_getsockopt(p->fd, level, -1, &v, &len);
	print_macro(p->test_number, p->test_value, p->test_name,
		    ret, -1, ENOPROTOOPT);
}

static void *test_exs_getsockopt(void *test_void_ptr)
{
	param *p = (param *)test_void_ptr;
	socklen_t len;
	uint32_t v32, e32;
	int test = p->test_number;
	int v, ret;

	switch (test) {
	case 1:
		p->test_value = "EPERM";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		len = sizeof(v);
		ret = exs_getsockopt(p->fd, SOL_SOCKET, SO_REUSEADDR,
				     &v, &len);
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EPERM);
		break;
	case 2:
		p->test_value = "EBADF";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v);
		ret = exs_getsockopt(-1, SOL_SOCKET, SO_REUSEADDR,
				     &v, &len);
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EBADF);
		break;
	case 3:
		p->test_value = "EOPNOTSUPP";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v);
		ret = exs_getsockopt(p->fd, -1, SO_REUSEADDR,
				     &v, &len);
		print_macro(p->test_number, p->test_value, p->test_name,
			    ret, -1, EOPNOTSUPP);
		break;
	case 4:
		test_getsockopt_enoprotoopt(p, SOL_SOCKET);
		break;
	case 5:
		test_getsockopt_enoprotoopt(p, SOL_EXS);
		break;
	case 6:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v);
		ret = exs_getsockopt(p->fd, SOL_SOCKET, SO_REUSEADDR,
				     &v, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v, 0, len, sizeof(int));
		break;
	case 7:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v32);
		errno = 0;
		ret = exs_getsockopt(p->fd, SOL_EXS, EXS_INLINE,
				     &v32, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v32, INT_MAX, len, sizeof(v32));
		break;
	case 8:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v) - 1;
		v = 0;
		ret = exs_getsockopt(p->fd, SOL_SOCKET, SO_REUSEADDR,
				     &v, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v, 0, len, sizeof(v));
		break;
	case 9:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v) + 1;
		ret = exs_getsockopt(p->fd, SOL_SOCKET, SO_REUSEADDR,
				     &v, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v, 0, len, sizeof(v));
		break;
	case 10:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v32) - 1;
		e32 = 0;
		v32 = INT_MAX;
		memcpy(&e32, &v32, len);
		v32 = 0;
		ret = exs_getsockopt(p->fd, SOL_EXS, EXS_INLINE,
				     &v32, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v32, e32, len, sizeof(v32));
		break;
	case 11:
		p->test_value = "OK_EXS_GETSOCKOPT";
		p->test_name = all_getsockopt_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		len = sizeof(v32) + 1;
		ret = exs_getsockopt(p->fd, SOL_EXS, EXS_INLINE,
				     &v32, &len);
		print_getsockopt_macro(p->test_number, p->test_value,
				       p->test_name, ret, errno,
				       v32, INT_MAX, len, sizeof(v32));
		break;
	default:
		printf("Invalid Test Number: %d\n", test);
		break;
	}

	return NULL;
} /* test_exs_getsockopt */

int main(int argc, char *argv[])
{
	execute(argc, argv, test_exs_getsockopt, all_getsockopt_tests, GETSOCKOPT_TESTS);
} /* main */
